/*
 * gawda.c - drawing area functions
 * 
 * include LICENSE
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <gtk/gtk.h>
#include <string.h>
#include <math.h>

#include <gaw.h>
 
#ifdef TRACE_MEM
#include <tracemem.h>
#endif
        


/*
 * simply forces widget w to redraw itself
 *    indirect call to expose event
 */
void da_drawing_redraw(GtkWidget *w)
{
   GdkRectangle rect;

   gtk_widget_get_allocation (w, &rect);
   rect.x = 0;
   rect.y = 0;

   if ( ! gtk_widget_get_realized (w) ) {
      return;
   }
   /* ask redraw */
   gdk_window_invalidate_rect (gtk_widget_get_window (w),
                               &rect, FALSE);
}

/* 
 * Set the X pointer cursor for all wavepanels: used to provide a
 * hint that we're expecting the user to drag out a line or region.
 */

void da_set_gdk_cursor(GtkWidget *w, int cursorType)
{
   GdkCursor *cursor;

   if (cursorType == -1) {
      cursor = NULL;
   } else {
      cursor = gdk_cursor_new(cursorType);
   }
   gdk_window_set_cursor(gtk_widget_get_window (w), cursor);

   if (cursor) {
      gdk_cursor_unref(cursor);
   }
}

/*
 * draw (or undraw) a vertical line used as cursor at offset x
 */
void da_draw_cursor(GtkWidget *widget, GdkGC *gc, int x)
{
   GdkRectangle rect;
   GdkWindow *window;

   gtk_widget_get_allocation (widget, &rect);
   window = gtk_widget_get_window (widget);
   if ( window ){
      gdk_draw_line(window, gc , x, 0, x, rect.height);
   }
}

void da_background_color_set(GtkWidget *widget, GdkColor *color)
{
   gtk_widget_modify_bg (widget, GTK_STATE_NORMAL, color);
}

void da_draw_srange(SelRange *sr)
{
   GdkWindow *window;

   window = gtk_widget_get_window (sr->wp->drawing);

   g_assert(sr->gc != NULL);
   if (sr->type & SR_X) {
      gdk_draw_line(window, sr->gc,
		    sr->x1, sr->y1, sr->x2, sr->y1);
   }
   if (sr->type & SR_Y) {
      gdk_draw_line(window, sr->gc,
		    sr->x1, sr->y1, sr->x1, sr->y2);
   }
   if (sr->type == SR_XY) {
      gdk_draw_line(window, sr->gc,
		    sr->x1, sr->y2, sr->x2, sr->y2);
      gdk_draw_line(window, sr->gc,
		    sr->x2, sr->y1, sr->x2, sr->y2);
   }
}

void
da_update_srange(SelRange *sr,  GdkEventMotion *event, int draw)
{
   int newx2, newy2;

   /*
    * the event->y does goofy things if the motion continues
    * outside the window, so we generate our own from the root
    * coordinates.
    */
   newx2 = event->x;
   newy2 = sr->y1 + (event->y_root - sr->y1_root);

   if (sr->drawn) {   /* undraw old */
      da_draw_srange(sr);
   }
   sr->drawn = draw;
   if (sr->type & SR_X) {
      sr->x2 = newx2;
   }
   if (sr->type & SR_Y) {
      sr->y2 = newy2;
   }
   if (draw) {       /* draw new if requested */
      da_draw_srange(sr);
   }
   msg_dbg( "type=%d newx=%d newy=%d draw=%d",
	   sr->type, sr->x2, sr->y2, draw);
   msg_dbg( "m %d %d %d %d",
	   (int) event->x, (int) event->y, 
	   (int) event->x_root, (int) event->y_root);
}

/*
 * done selecting range; do the callback
 */

void
da_callback_srange(UserData *ud, WavePanel *wp )
{
   SelRange *sr = ud->srange;
   double xstart, ystart;
   double xend, yend;
   GawLabels *lbx = ud->xLabels;
   GawLabels *lby = sr->wp->yLabels;
   
   msg_dbg( "type=%d x1=%d x2=%d  y1=%d y2=%d",
	   sr->type, sr->x1, sr->x2, sr->y1, sr->y2);

   xstart = al_label_x2val(lbx, sr->x1);
   xend   = al_label_x2val(lbx, sr->x2);

   ystart = al_label_y2val(lby, sr->y1 );
   yend   = al_label_y2val(lby, sr->y2 );
 
   if ( ystart < yend ) {
      pa_panel_set_yvals( wp, ystart, yend);
   } else {
      pa_panel_set_yvals( wp, yend, ystart);
   }

   switch (sr->type) {
    case SR_X:
      az_cmd_zoom_absolute(ud, xstart, xend );
      break;
    case SR_Y:
      wp->man_yzoom = 1;
      break;
    case SR_XY:
      wp->man_yzoom = 1;
      az_cmd_zoom_absolute(ud, xstart, xend );
      break;
   }
}


/*
 * drawing area configure handler
 * The "configure_event" signal to take any necessary actions
 *   when the widget changes size.
 */
static gboolean
da_drawing_configure_cb (GtkWidget *widget, GdkEventConfigure *event,
		      gpointer data )
{
   WavePanel *wp =  (WavePanel *) data;
   GtkAllocation walloc;
   GdkWindow *window = gtk_widget_get_window (widget);
   
   gtk_widget_get_allocation (widget, &walloc);
   msg_dbg( "w %d h = %d new %d %d 0x%lx",
	    walloc.width,
	    walloc.height,
	    event->width,  event->height,
	    (long unsigned int) widget );

   if ( ! wp->ud->bg_color ) {
      ac_color_initialize(wp);
   }
   if ( ! wp->grid_gc ) {
      ac_color_panel_grid_init(wp);
   }
   if ( wp->pixmap ){
      g_object_unref (wp->pixmap);
   }
   
   wp->pixmap = gdk_pixmap_new (window,
                           walloc.width,
                           walloc.height, -1);

   /* Initialize the pixmap to bg color */
   GtkStyle *style = gtk_widget_get_style (widget);
   gdk_draw_rectangle (wp->pixmap,
		       style->bg_gc[gtk_widget_get_state(widget)],
		       TRUE,  0, 0,
		       walloc.width,
		       walloc.height );

   pa_panel_label_size(wp);
   wp->configure_seen = 1;
   return TRUE;
}


/*
 *  draw one wave in a panel - gwave algo modified
 *  sweep from x and calculate  y point
 */
void da_drawing_draw_wave_algo (VisibleWave *vw, WavePanel *wp)
{
   int x0 = -1;
   int x1;
   int y0 = 0;
   int y1;
   int i;
   double xval;
   double yval;
   GawLabels *lbx = wp->ud->xLabels;
   GawLabels *lby = wp->yLabels;
   
   for ( i = 0 ; i < lbx->w ; i++ ) {
      xval = al_label_x2val(lbx, i);
      yval = wavevar_get_yvalue(vw->var, xval, lbx->npoints);
      x1 = i;
      y1 = al_label_val2y(lby, yval);
//      fprintf(stderr, "i %d, x0 = %d, y0 = %d, x1 = %d, y1 = %d\n",
//              i, x0, y0,  x1, y1);
      if ( x0 >= 0 ){
         gdk_draw_line(wp->pixmap, vw->gc, x0, y0,  x1, y1);
      }
      x0 = x1;
      y0 = y1;
   }
}

#if 0
/*
 *  draw one wave in a panel - gwave algo original
 *  sweep from x and calculate  y point
 *
 * to be removed
 */
void da_drawing_draw_wave_algo0 (VisibleWave *vw, WavePanel *wp)
{
   int x0, x1;
   int y0, y1;
   int i;
   double xstep;
   double xval;
   double yval;
   double xmin = wavevar_ivar_get_min(vw->var) ;
   double xmax = wavevar_ivar_get_max(vw->var) ;
   int w = wp->drawing->allocation.width;
   int h = wp->drawing->allocation.height;
   GawLabels *lbx = wp->ud->xLabels;
   GawLabels *lby = wp->yLabels;
   
   xstep = (lbx->end_val - lbx->start_val) / w;  /* linear only */

   xval = lbx->start_val;
   yval = wavevar_get_yvalue(vw->var, xval, lbx->npoints);
   x1 = 0;
   if ( al_label_do_logAxis(lby) ) {
      y1 = VAL2LY(lby, yval, h);
   } else {
      y1 = VAL2Y(lby, yval, h);
   }

   for ( i = 0 ; i < w ; i++ ) {
      x0 = x1;
      y0 = y1;
      x1 = x0 + 1;
      if ( al_label_do_logAxis(lbx) ) {
	 xval = X2LVAL(lbx, x1, w);
      } else {
	 xval += xstep;
      }
      if (xmin <= xval && xval <= xmax ) {
	 yval = wavevar_get_yvalue(vw->var, xval, lbx->npoints);
	 if ( al_label_do_logAxis(lby) ) {
	    y1 = VAL2LY(lby, yval, h);
	 } else {
	    y1 = VAL2Y(lby, yval, h);
	 }
	 gdk_draw_line(wp->pixmap, vw->gc, x0, y0,  x1, y1);
      }
   }
}
#endif
/*
 *  draw one wave in a panel
 */
void da_drawing_draw_wave (VisibleWave *vw, WavePanel *wp)
{
   g_assert(vw->gc != NULL);
   if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(vw->button))) {
      gdk_gc_set_line_attributes(vw->gc, 2, GDK_LINE_SOLID,
				 GDK_CAP_BUTT, GDK_JOIN_ROUND);
   } else {
      gdk_gc_set_line_attributes(vw->gc, 1, GDK_LINE_SOLID,
				 GDK_CAP_BUTT, GDK_JOIN_ROUND);
   }

   da_drawing_draw_wave_algo (vw, wp);
}

/*
 *  draw grid in a panel
 */
static void
da_drawing_draw_grid (WavePanel *wp)
{
   GtkAllocation walloc;
   gtk_widget_get_allocation (wp->drawing, &walloc);
   int w = walloc.width;
   int h = walloc.height;

   ArrayStruct *ary = array_struct_new( sizeof(GdkSegment), 128, NULL);
   GawLabels *lbx = wp->ud->xLabels;
   GawLabels *lby = wp->yLabels;

   /* calculate grid segments */
   
   /* vertical lines */
   if ( al_label_do_logAxis( lbx) ) {
      al_label_draw_vlog_grid( lbx, ary, w, h );
   } else {
      al_label_draw_vlin_grid( lbx, ary, w, h );
   }

   /* horizontal lines */
   if ( al_label_do_logAxis( lby) ) {
      al_label_draw_hlog_grid( lby, ary, w, h );
   } else {
      al_label_draw_hlin_grid( lby, ary, w, h );
   }

   /* draw grid */
   gdk_draw_segments(wp->pixmap, wp->grid_gc,
                     (GdkSegment *) array_struct_table_get(ary),
                     array_struct_get_nelem(ary) );
   array_struct_destroy(ary);
}

/*
 * drawing area expose handler
 *  draw all the panel
 */
gboolean
da_drawing_expose_cb (GtkWidget *widget, GdkEventExpose *event, gpointer data )
{
   WavePanel *wp =  (WavePanel *) data;
   UserData  *ud =   wp->ud;
   GawLabels *lby = wp->yLabels;
   GtkAllocation walloc;
   gtk_widget_get_allocation (widget, &walloc);
   
   int w = walloc.width;
   int h = walloc.height;
   int i;
   int y;

   msg_dbg( "called width %d height %d 0x%lx", w, h,
	    (long unsigned int) widget );
   if ( wp->configure_seen ) {
      wp->configure_seen = 0;
      /*
       * Need this here for getting data when they are valid for :
       *    - init purpose
       *    - set the size of label box :
       *         - after a main window change
       *         - main scroll bar show/hide
       */
      if ( (lby->changed & CV_INIT) == 0){ /* panel creation */
	 lby->changed |= CV_INIT;
	 pa_panel_update_min_max(wp);
      }
      if ( lby->changed & CV_CHANGED  ){
	 lby->changed &= ~CV_CHANGED ;
	 pa_panel_draw_ylabels( wp );
      }
      
      GawLabels *lbx = wp->ud->xLabels;
      if ( lbx->changed & (CV_CHANGED | CV_SBCHANGED) ){
	 int vis = lbx->changed & CV_SBSHOW ;
	 int xlwidth = w;

	 if ( vis ) {
	    xlwidth += ud->sbSize;
	 }
	 if ( xlwidth != lbx->wh ){ 
	    gtk_widget_set_size_request (GTK_WIDGET(ud->xlabel_box),
					 xlwidth, lbx->lbheight);
	    lbx->wh = xlwidth; /* x label box width */
	    lbx->changed |= CV_CHANGED ;
	 }
	 lbx->changed &= ~CV_SBCHANGED ;
	 msg_dbg("w %d, h %d, lbx->wh %d, visible %d", lbx->w, lbx->h, lbx->wh, vis);
      }
      if ( (lbx->changed & CV_INIT) == 0){
	 /* init data at first run */
	 pa_panel_update_all_data(ud); /* this call al_label_draw */
	 lbx->changed |= CV_INIT | CV_CHANGED;
	 lbx->changed &= ~CV_CHANGED;
      }
      if ( lbx->changed & CV_CHANGED ){
	 lbx->changed &= ~CV_CHANGED ;
	 al_label_draw( lbx );
      }

      ud->panelHeight = h;
      ud->panelWidth = w;
//      msg_dbg("HQ1 w %d, h %d", ud->panelWidth, ud->panelHeight);
      GtkWidget *scr = gtk_scrolled_window_get_vscrollbar(
                               GTK_SCROLLED_WINDOW(ud->panel_scrolled)) ;
      if ( gtk_widget_get_visible (scr) == FALSE ) {
	 ud->panelWidth = w - ud->sbSize ;
      }
      if ( ud->panelWidth % 2 ) {
	 ud->panelWidth += 1;
      }
//      msg_dbg("HQ2 w %d, h %d, sb %d", ud->panelWidth, ud->panelHeight, ud->sbSize);
   }
   if ( ud->suppress_redraw ) {
      return FALSE;
   }

   /* clear all */
   GtkStyle *style = gtk_widget_get_style (widget);
   gdk_draw_rectangle(wp->pixmap, 
		      style->bg_gc[gtk_widget_get_state(widget)],
		      TRUE, 0, 0,  w, h);

   if (wp->selected) {
      gdk_draw_line(wp->pixmap, ud->hl_gc,     1, 1,      w - 2, 1);
      gdk_draw_line(wp->pixmap, ud->hl_gc, w - 2, 1,  w - 2, h - 2);
      gdk_draw_line(wp->pixmap, ud->hl_gc, w - 2, h - 2,  1, h - 2);
      gdk_draw_line(wp->pixmap, ud->hl_gc,     1, h - 2,  1,     1);
   }
   if ( wp->showGrid ) {
      /* graticule */
      da_drawing_draw_grid (wp);
   } else {
      /* draw horizontal line at y=zero. */
      if (lby->start_val < 0 && lby->end_val > 0) {
	 y = VAL2Y(lby, 0, h);
	 gdk_draw_line(wp->pixmap, wp->grid_gc, 0, y, w, y);
      }
   }
   /* draw waves */
   g_list_foreach(wp->vwlist, (GFunc) da_drawing_draw_wave, wp); 

   /* draw the 2 cursors in the panel */
   for (i = 0 ; i < 2 ; i++) {                        
      AWCursor *csp = ud->cursors[i];
      if (csp->shown) {
	 gdk_draw_line(wp->pixmap, csp->cursor_gc, csp->x, 0, csp->x, h);
      }
   }

   /* draw select-range line, if in this WavePanel */
   if ( ud->srange && ud->srange->drawn && ud->srange->wp == wp) {
      da_draw_srange(ud->srange);
   }

   GdkWindow *window = gtk_widget_get_window (widget);
   if (event) {
      /* Draw the exposed portions of the pixmap in its window. */
      gdk_draw_drawable (window,
			 style->fg_gc[gtk_widget_get_state(widget)],
			 wp->pixmap,
			 event->area.x, event->area.y,
			 event->area.x, event->area.y,
			 event->area.width, event->area.height);
   } else {
      /* Draw the whole thing. */
      gdk_draw_drawable (window,
			 style->fg_gc[gtk_widget_get_state(widget)],
			 wp->pixmap,
			 0, 0, 0, 0, w, h);
   }
      
   return FALSE;
}

char *da_statusFormat = N_("Panel W %d H %d X %d Y %d");

/*
 * drawing area mouse button press handler
 */
static gboolean
da_drawing_button_press_cb (GtkWidget *widget, GdkEventButton *event,
			 gpointer data )
{
   WavePanel *wp =  (WavePanel *) data;
   UserData  *ud =   wp->ud;
   
   msg_dbg( "button %d state %d mouseState %d",
	   event->button, event->state, ud->mouseState  );
   if ( wp->pixmap == NULL){
      return FALSE; /* we haven't gotten a configure event */
   }
   if ( event->button == 3 ){
      if ( wp->selected ) {
	 /* unselect all */
	 pa_panel_set_selected( NULL, ud );
	 da_drawing_redraw(wp->drawing);
      }

      GtkWidget *menu = gtk_ui_manager_get_widget (wp->ui, "/PanelPopup");
      g_object_set_data (G_OBJECT(menu), "PanelPopup-action", (gpointer) wp);
      
      gtk_menu_popup (GTK_MENU (menu), NULL, NULL,
		      NULL, NULL,  3, event->time);

      return TRUE;
   }
   char text[128];

   if ( event->button == 1 ) {
      GtkAllocation walloc;
      gtk_widget_get_allocation (widget, &walloc);

      sprintf (text, gettext(da_statusFormat), walloc.width,
	       walloc.height, (int) event->x, (int) event->y );
      gtk_statusbar_push (GTK_STATUSBAR(ud->statusbar), 2, text);
   }
   switch (ud->mouseState) {
    case M_NONE:
      pa_panel_set_selected( wp, ud );
      
      gtk_grab_add(widget);
      ud->mouseState = M_CURSOR_DRAG;
      ud->button_down = event->button;

      da_set_gdk_cursor(widget, GDK_SB_H_DOUBLE_ARROW);
      ud->drag_button = event->button;
      cu_display_xcursor(wp, ud->drag_button, event->x);

      break;

    case M_SELRANGE_ARMED:
      gtk_grab_add(widget);
      ud->button_down = event->button;
      ud->mouseState = M_SELRANGE_ACTIVE;

      ud->srange->y1 = ud->srange->y2 = event->y;
      ud->srange->x1 = ud->srange->x2 = event->x;
      ud->srange->x1_root = event->x_root;
      ud->srange->y1_root = event->y_root;
      ud->srange->wp = wp;

      break;
      /* can't start another drag until first one done */

    case M_CURSOR_DRAG:
      break;
    case M_SELRANGE_ACTIVE:
    default:
      break;
   }
   ap_all_panel_redraw(ud);

   return TRUE;
}
/*
 * drawing area mouse button release handler
 */
static gboolean
da_drawing_button_release_cb (GtkWidget *widget, GdkEventButton *event,
			   gpointer data )
{
   WavePanel *wp =  (WavePanel *) data;
   UserData  *ud =  wp->ud;

   app_memcheck();
   msg_dbg( "button %d state %d", event->button, ud->mouseState );
   if ( wp->pixmap == NULL || ud->button_down != event->button ){
      return FALSE; /* we haven't gotten a configure event */
   }
   if ( event->button == 1 ) {
      gtk_statusbar_pop (GTK_STATUSBAR(ud->statusbar), 2);
   }
   GdkWindow *window = gtk_widget_get_window (widget);
   switch(ud->mouseState) {
    case M_CURSOR_DRAG:
      gtk_grab_remove(widget);
      gdk_window_set_cursor(window, NULL);
      cu_display_xcursor(wp, ud->drag_button, event->x);
      ud->drag_button = -1;
      break;

    case M_SELRANGE_ACTIVE:
      gtk_grab_remove(widget);
      g_list_foreach(ud->panelList, (GFunc) pa_panel_drawing_set_gdk_cursor,
		     GINT_TO_POINTER (-1) ); /* clear gdk cursor */
      da_update_srange(ud->srange, (GdkEventMotion *) event, 0);
      da_callback_srange(ud, wp);
      break;

    default:
      break;
   }
   ud->mouseState = M_NONE;
   pa_panel_full_redraw(wp);
   return TRUE;
}

/*
 * drawing area mouse motion handler
 */
static gboolean
da_drawing_motion_cb (GtkWidget *widget, GdkEventMotion *event, gpointer data )
{
   WavePanel *wp =  (WavePanel *) data;
   UserData  *ud =  wp->ud;
   char text[128];
   
   msg_dbg( "state %d", ud->mouseState );
   if ( wp->pixmap == NULL){
      return FALSE; /* we haven't gotten a configure event */
   }

   GtkAllocation walloc;
   gtk_widget_get_allocation (widget, &walloc);
   sprintf (text, gettext(da_statusFormat), walloc.width,
	    walloc.height, (int) event->x, (int) event->y );
   gtk_label_set_text ( GTK_LABEL(ud->statusLabel), text); 

   switch(ud->mouseState) {
    case M_CURSOR_DRAG:
      cu_display_xcursor(wp, ud->drag_button, event->x);
      break;

    case M_SELRANGE_ACTIVE:
      da_update_srange(ud->srange, event, 1);
      break;

    default:
      break;
   }

   return TRUE;
}

/*
 * da are created to their minimal size
 * They automtically expand to the main window allocated size.
 */

void da_drawing_set_size_request(GtkWidget *drawing, int w, int h)
{
   gtk_widget_set_size_request (GTK_WIDGET(drawing),
                         w, h );
   msg_dbg( "w %d, h %d", w, h );
}


/*
 * Construct drawing area.
 */ 
 GtkWidget *da_drawing_create( WavePanel *wp )
{
   UserData *ud = wp->ud;
   UserPrefs *up = ud->up;
   GtkWidget *drawing;
   
   msg_dbg("width = %d, height = %d, showXlabels = %d",
	   up->panelWidth, up->panelHeight, up->showXLabels);

   /* drawing area for waveform */
   drawing = gtk_drawing_area_new();
   gtk_widget_set_name( drawing, "wavepanel");
   g_object_ref (drawing); /* increment ref to avoid destruction */

   g_signal_connect( drawing, "expose_event", 
		     G_CALLBACK (da_drawing_expose_cb), (gpointer) wp);
   g_signal_connect (drawing,"configure_event",
		     G_CALLBACK (da_drawing_configure_cb), (gpointer) wp);

   g_signal_connect( drawing, "button_press_event", 
		     G_CALLBACK (da_drawing_button_press_cb), (gpointer) wp);
   g_signal_connect( drawing, "button_release_event", 
		     G_CALLBACK (da_drawing_button_release_cb), (gpointer) wp);
   g_signal_connect( drawing, "motion_notify_event", 
		     G_CALLBACK ( da_drawing_motion_cb), (gpointer) wp);

   da_drawing_set_size_request(drawing, up->minPanelWidth, up->minPanelHeight );
   
   gtk_widget_show(drawing);
   if ( ud->bg_color ) {
      da_background_color_set(drawing, ud->bg_color);
   }
   
   /* Set up a drawing as a drop target */
   ad_set_drag_dest(drawing, ud, wp, DND_PANEL);

   gtk_widget_set_events(drawing, GDK_EXPOSURE_MASK |
			 GDK_BUTTON_RELEASE_MASK |
			 GDK_BUTTON_PRESS_MASK |
			 GDK_BUTTON1_MOTION_MASK |
			 GDK_BUTTON2_MOTION_MASK);
   return drawing;
}

